/* 
 **  mod_memcached.c -- Apache module for put/delete/head/get
 */ 

#include "httpd.h"
#include "apr.h"
#include "apr_strings.h"
#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_request.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"
#include "version.h"
#include <libmemcached/memcached.h>
#include "watchpoint.h"

#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>

#include <assert.h>

module AP_MODULE_DECLARE_DATA memcached_module;

typedef struct {
  int delete;
  int put;
  int get;
  int overwrite;
  char *servers;
  char *size_limit;
  char *expiration;
} memcached_conf;

void
set_config_defaults(apr_pool_t *p, memcached_conf *cfg)
{
  bzero(cfg, sizeof(memcached_conf));
}

static void *create_dir_mconfig(apr_pool_t *p, char *dir) 
{
  memcached_conf *cfg;

  cfg= apr_pcalloc(p, sizeof(memcached_conf));
  set_config_defaults(p, cfg);

  return (void *) cfg;
}

static void *merge_dir_mconfig(apr_pool_t *p, void *origin, void *new) 
{
  memcached_conf *cfg_origin= (memcached_conf *)origin;
  memcached_conf *cfg_new= (memcached_conf *)new;

  return (void *) ( cfg_new ? cfg_new : cfg_origin );
}

char * location(request_rec *r)
{
  char *key;

  key= (char *)rindex(r->uri, '/');

  if (key)
  {
    key++;

    if (*key)
      return key;
    else
      return NULL;
  }
  else
    return NULL;
}

int get_method(request_rec *r, memcached_st *memc)
{
  memcached_return rc;
  memcached_result_st result;
  char *key;
  size_t key_length;
  int return_code;

  if ((return_code = ap_discard_request_body(r)) != OK)
    return return_code;

  key= location(r);

  /* If we are only after the header or the return_code is null we can just bail */
  if (r->header_only || key == NULL)
    return key ? HTTP_OK : HTTP_NOT_FOUND;

  key_length= strlen(key);
  rc= memcached_mget(memc, &key, &key_length, 1);

  if (rc != MEMCACHED_SUCCESS)
    return HTTP_NOT_FOUND;

  if (memcached_result_create(memc, &result) == NULL)
    return HTTP_NOT_FOUND;


  memcached_fetch_result(memc, &result, &rc);
    return HTTP_NOT_FOUND;

  if (rc != MEMCACHED_SUCCESS)
    return_code= HTTP_NOT_FOUND;
  else
  {
    request_rec *subr;
    uint64_t cas;
    char buffer[1024];

    subr = (request_rec *) ap_sub_req_lookup_file(key, r, NULL);
    ap_set_content_type(r, subr->content_type);
    r->content_type = apr_pstrdup(r->pool, subr->content_type);
    r->handler = apr_pstrdup(r->pool, subr->handler);
    WATCHPOINT_STRING(subr->content_type);
    WATCHPOINT_STRING(subr->handler);
    ap_destroy_sub_req(subr);

    cas= memcached_result_cas(&result);
    sprintf(buffer, "%llu", (unsigned long long)cas);
    apr_table_setn(r->headers_out, "X-CAS", buffer);

    return_code= OK;
    ap_rwrite(memcached_result_value(&result), memcached_result_length(&result), r);
  }

  memcached_result_free(&result);

  return return_code;
}

int delete_method(request_rec *r, memcached_st *memc)
{
  memcached_return rc;
  char *key;
  const char *tmp_string;
  time_t expiration= 0;
  int result;

  if ((result = ap_discard_request_body(r)) != OK)
    return result;

  if ((key= location(r)) == NULL)
    return HTTP_NOT_FOUND;

  tmp_string=  apr_table_get(r->headers_in, "X-Expiration-Seconds");

  if (tmp_string)
  {
    expiration= (time_t)strtol(tmp_string, (char **) NULL, 10);
  }

  rc= memcached_delete(memc, key, strlen(key), expiration);

  if (rc == MEMCACHED_SUCCESS)
    result= OK;
  else
    result= HTTP_NOT_FOUND;

  return result;
}

int put_method(request_rec *r, memcached_st *memc, 
               memcached_conf *cfg,
               uint32_t size_limit, time_t expiration)
{
  memcached_return rc;
  char *key;
  const char *tmp_string;
  int rv, seen_eos;
  apr_bucket_brigade *bb;
  uint64_t cas;
  memcached_string_st buffer;


  if ((key= location(r)) == NULL)
    return HTTP_CONFLICT;

  bb= apr_brigade_create(r->pool, r->connection->bucket_alloc);
  seen_eos = 0;

  memcached_string_create(memc, &buffer, HUGE_STRING_LEN);

  do {
    apr_bucket *bucket;

    rv= ap_get_brigade(r->input_filters, bb, AP_MODE_READBYTES,
                       APR_BLOCK_READ, HUGE_STRING_LEN);

    if (rv != APR_SUCCESS) 
    {
      if (rv == AP_FILTER_ERROR) 
      {
        apr_brigade_destroy(bb);
        return rv;
      }
      else 
      {
        apr_brigade_destroy(bb);
        return HTTP_BAD_REQUEST;
      }
    }

    for (bucket= APR_BRIGADE_FIRST(bb);
         bucket != APR_BRIGADE_SENTINEL(bb);
         bucket= APR_BUCKET_NEXT(bucket)) 
    {
      const char *data;
      apr_size_t len;

      if (APR_BUCKET_IS_EOS(bucket)) 
      {
        seen_eos = 1;
        break;
      }

      if (bucket->length == 0)
        continue;

      rv= apr_bucket_read(bucket, &data, &len, APR_BLOCK_READ);
      if (rv != APR_SUCCESS) 
      {
        apr_brigade_destroy(bb);
        return HTTP_BAD_REQUEST;
      }

      memcached_string_append(&buffer, (char *)data, len);
    }
    apr_brigade_cleanup(bb);
  } while (!seen_eos);

  tmp_string=  apr_table_get(r->headers_in, "X-CAS");
  cas= tmp_string ? strtol(tmp_string, (char **) NULL, 10) : 0;

  tmp_string=  apr_table_get(r->headers_in, "X-Expiration-Seconds");

  if (tmp_string)
  {
    uint32_t seconds;

    seconds= strtol(tmp_string, (char **) NULL, 10);

    if (expiration)
      expiration= seconds > expiration ? expiration : seconds;
    else 
      expiration= seconds;
  }

  if (size_limit == 0 || size_limit >= memcached_string_length(&buffer))
  {
    if (cas)
    {
      rc= memcached_cas(memc, key, strlen(key), 
                        memcached_string_value(&buffer), 
                        memcached_string_length(&buffer),
                        expiration, (uint32_t)0,
                        cas);
    }
    else
    {
      if (cfg->overwrite)
        rc= memcached_set(memc, key, strlen(key), 
                          memcached_string_value(&buffer), 
                          memcached_string_length(&buffer),
                          expiration, (uint32_t)0);
      else
        rc= memcached_add(memc, key, strlen(key), 
                          memcached_string_value(&buffer), 
                          memcached_string_length(&buffer),
                          expiration, (uint32_t)0);

      memcached_string_free(&buffer);
    }

    return rc == MEMCACHED_SUCCESS ? OK : DECLINED;
  }

  memcached_string_free(&buffer);

  return DECLINED;
}

/* The sample content handler */
static int memcached_handler(request_rec *r)
{
  memcached_conf *cfg= ap_get_module_config (r->per_dir_config, &memcached_module);
  memcached_st *memc;
  memcached_server_st *servers;
  memcached_return rc;
  int result;

  apr_table_setn(r->headers_out, "X-Powered-By", "ModMemcached/"VERSION);

  memc= memcached_create(NULL);
  servers= memcached_servers_parse(cfg->servers);
  rc= memcached_server_push(memc, servers);
  memcached_behavior_set(memc, MEMCACHED_BEHAVIOR_SUPPORT_CAS, 1);
  memcached_behavior_set(memc, MEMCACHED_BEHAVIOR_VERIFY_KEY, 1);

  switch (r->method_number)
  {
  case M_DELETE:
    if (cfg->delete == 0)
      result= DECLINED;
    else /* We discard the body here so that we can resuse delete_method for update */
    {
      ap_discard_request_body(r);
      result= delete_method(r, memc);
    }
    break;
  case M_PUT:
    if (cfg->put == 0)
      result= DECLINED;
    else
    {
      uint32_t size_limit;
      time_t expiration;

      size_limit= cfg->size_limit ? strtol(cfg->size_limit, (char **) NULL, 10) : 0;
      expiration= cfg->expiration ? strtol(cfg->expiration, (char **) NULL, 10) : 0;

      result= put_method(r, memc, cfg, size_limit, expiration);
    }
    break;
  case M_GET:
    if (cfg->get == 0)
      result= DECLINED;
    else
      result= get_method(r, memc);
    break;
  default:
    result= HTTP_METHOD_NOT_ALLOWED;
    break;
  }

  memcached_server_list_free(servers);
  memcached_free(memc);

  return result;
}

static void memcached_register_hooks(apr_pool_t *p)
{
  ap_hook_handler(memcached_handler, NULL, NULL, APR_HOOK_FIRST);
}

static const command_rec memcached_cmds[]= {
  AP_INIT_FLAG("Memcached", ap_set_flag_slot, (void *) APR_OFFSETOF(memcached_conf, get), OR_ALL, 
               "On or Off, controls if this module is enabled or not (default off)"),
  AP_INIT_FLAG("MemcachedDelete", ap_set_flag_slot, (void *) APR_OFFSETOF(memcached_conf, delete), OR_ALL, 
               "On or Off, controls if this module is enabled or not (default off)"),
  AP_INIT_FLAG("MemcachedPut", ap_set_flag_slot, (void *) APR_OFFSETOF(memcached_conf, put), OR_ALL, 
               "On or Off, controls if this module is enabled or not (default off)"),
  AP_INIT_FLAG("MemcachedOverwrite", ap_set_flag_slot, (void *) APR_OFFSETOF(memcached_conf, overwrite), OR_ALL, 
               "Whether or not a PUT will overwrite and existing object or not."),
  AP_INIT_TAKE1("MemcachedServers", ap_set_string_slot, (void *) APR_OFFSETOF(memcached_conf, servers), OR_ALL, 
                "List of memcached servers to use (in server:port format)"),
  AP_INIT_TAKE1("MemcachedSizeLimit", ap_set_string_slot, (void *) APR_OFFSETOF(memcached_conf, size_limit), OR_ALL, 
                "Limit on the size of an object that can be posted to the server."),
  AP_INIT_TAKE1("MemcachedExpiration", ap_set_string_slot, (void *) APR_OFFSETOF(memcached_conf, expiration), OR_ALL, 
                "Expiration time placed on items which are placed into memcached"),
  {NULL}
};

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA memcached_module = {
  STANDARD20_MODULE_STUFF, 
  create_dir_mconfig,    /* create per-dir    config structures */
  merge_dir_mconfig,     /* merge  per-dir    config structures */
  NULL,                  /* create per-server config structures */
  NULL,                  /* merge  per-server config structures */
  memcached_cmds,                  /* table of config file commands       */
  memcached_register_hooks  /* register hooks                      */
};

